import os
import cPickle
import utils
import numpy as np
from scipy.stats import binom, scoreatpercentile
from scipy.optimize import curve_fit
from scipy import exp
import operator
from copy import copy, deepcopy
from collections import defaultdict, Counter
import re
from pyteomics import parser, mass, fasta, auxiliary as aux, achrom
try:
    from pyteomics import cmass
except ImportError:
    cmass = mass
import subprocess
from sklearn import linear_model
import tempfile
import pandas as pd
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
try:
    import seaborn
    seaborn.set(rc={'axes.facecolor':'#ffffff'})
    seaborn.set_style('whitegrid')
except:
    pass


def get_RCs2(sequences, RTs, lcp = -0.21,
            term_aa = False, **kwargs):
    labels = kwargs.get('labels')

    # Make a list of all amino acids present in the sample.
    peptide_dicts = [
            parser.amino_acid_composition(peptide, False, term_aa,
                               allow_unknown_modifications=True,
                               labels=labels)
            if not isinstance(peptide, dict) else peptide
        for peptide in sequences]

    detected_amino_acids = {aa for peptide_dict in peptide_dicts
                                for aa in peptide_dict}

    composition_array = []
    for pdict in peptide_dicts:
        loglen = np.log(parser.length(pdict))
        composition_array.append([pdict.get(aa, 0.)
             * (1. + lcp * loglen)
               for aa in detected_amino_acids] + [1.])

    if term_aa:
        for term_label in ['nterm', 'cterm']:
            normalizing_peptide = []
            for aa in detected_amino_acids:
                if aa.startswith(term_label):
                    normalizing_peptide.append(1.0)
                elif (term_label+aa) in detected_amino_acids:
                    normalizing_peptide.append(-1.0)
                else:
                    normalizing_peptide.append(0.0)
            normalizing_peptide.append(0.0)
            composition_array.append(normalizing_peptide)
            RTs.append(0.0)
    
    model_ransac = linear_model.RANSACRegressor(linear_model.LinearRegression(n_jobs=12), min_samples=0.5, max_trials=5000, random_state=42)
    model_ransac.fit(np.array(composition_array), np.array(RTs))
    RCs = model_ransac.estimator_.coef_
    if term_aa:
        for term_label in ['nterm', 'cterm']:
            RTs.pop()

    # Form output.
    RC_dict = {}
    RC_dict['aa'] = dict(
        zip(list(detected_amino_acids),
            RCs[:len(detected_amino_acids)]))
    RC_dict['aa'][parser.std_nterm] = 0.0
    RC_dict['aa'][parser.std_cterm] = 0.0
    RC_dict['const'] = RCs[len(detected_amino_acids)]
    RC_dict['lcp'] = lcp

    # Find remaining terminal RCs.
    if term_aa:
        for term_label in ['nterm', 'cterm']:
            # Check if there are terminal RCs remaining undefined.
            undefined_term_RCs = [aa for aa in RC_dict['aa']
                                if aa[1:5] != 'term'
                                and term_label + aa not in RC_dict['aa']]
            if not undefined_term_RCs:
                continue

            # Find a linear relationship between internal and terminal RCs.
            defined_term_RCs = [aa for aa in RC_dict['aa']
                              if aa[1:5] != 'term'
                              and term_label + aa in RC_dict['aa']]

            a, b, r, stderr = linear_regression(
                [RC_dict['aa'][aa] for aa in defined_term_RCs],
                [RC_dict['aa'][term_label+aa] for aa in defined_term_RCs])

            # Define missing terminal RCs using this linear equation.
            for aa in undefined_term_RCs:
                RC_dict['aa'][term_label + aa] = a * RC_dict['aa'][aa] + b
    inlier_mask = model_ransac.inlier_mask_
    outlier_mask = np.logical_not(inlier_mask)
    return RC_dict, outlier_mask


def process_file(args):
    # fname = args['file']
    # ftype = fname.rsplit('.', 1)[-1].lower()
    utils.seen_target.clear()
    utils.seen_decoy.clear()
    return process_peptides(args)


def peptide_processor(peptide, **kwargs):
    seqm = peptide
    results = []
    m = cmass.fast_mass(seqm, aa_mass=kwargs['aa_mass']) + kwargs['aa_mass'].get('Nterm', 0) + kwargs['aa_mass'].get('Cterm', 0)
    acc_l = kwargs['acc_l']
    acc_r = kwargs['acc_r']
    dm_l = acc_l * m / 1.0e6
    dm_r = acc_r * m / 1.0e6
    start = nmasses.searchsorted(m - dm_l)
    end = nmasses.searchsorted(m + dm_r)
    idx = set(range(start, end))
    for i in idx:
        peak_id = ids[i]
        I = Is[i]
        massdiff = (m - nmasses[i]) / m * 1e6
        results.append((seqm, massdiff, rts[i], peak_id, I, Scans[i], Isotopes[i], mzraw[i], avraw[i], charges[i]))
    return results


def prepare_peptide_processor(fname, args):
    global nmasses
    global rts
    global charges
    global ids
    global Is
    global Scans
    global Isotopes
    global mzraw
    global avraw
    nmasses = []
    rts = []
    charges = []
    ids = []
    Is = []
    Scans = []
    Isotopes = []
    mzraw = []
    avraw = []

    min_ch = args['cmin']
    max_ch = args['cmax']

    min_isotopes = args['i']
    min_scans = args['sc']

    print 'Reading spectra ...'
    for m, RT, c, peak_id, I, nScans, nIsotopes, mzr, avr in utils.iterate_spectra(fname, min_ch, max_ch, min_isotopes, min_scans):
        nmasses.append(m)
        rts.append(RT)
        charges.append(c)
        ids.append(peak_id)
        Is.append(I)
        Scans.append(nScans)
        Isotopes.append(nIsotopes)
        mzraw.append(mzr)
        avraw.append(avr)

    i = np.argsort(nmasses)
    nmasses = np.array(nmasses)[i]
    rts = np.array(rts)[i]
    charges = np.array(charges)[i]
    ids = np.array(ids)[i]
    Is = np.array(Is)[i]
    Scans = np.array(Scans)[i]
    Isotopes = np.array(Isotopes)[i]
    mzraw = np.array(mzraw)[i]
    avraw = np.array(avraw)[i]

    fmods = args['fmods']
    aa_mass = mass.std_aa_mass
    if fmods:
        for mod in fmods.split(','):
            m, aa = mod.split('@')
            if aa == '[':
                aa_mass['Nterm'] = float(m)
            elif aa == ']':
                aa_mass['Cterm'] = float(m)
            else:
                aa_mass[aa] += float(m)

    acc_l = args['ptol']
    acc_r = args['ptol']

    return {'aa_mass': aa_mass, 'acc_l': acc_l, 'acc_r': acc_r, 'args': args}


def peptide_processor_iter_isoforms(peptide, **kwargs):
    out = []
    out.append(peptide_processor(peptide, **kwargs))
    return out

def get_results(ms1results):
    resdict = dict()
    labels = [
        'seqs',
        'md',
        'rt',
        'ids',
        'Is',
        'Scans',
        'Isotopes',
        'mzraw',
        'av',
        'ch'
    ]
    for label, val in zip(labels, zip(*ms1results)):
        resdict[label] = np.array(val)
    return resdict

def filter_results(resultdict, idx):
    for label in resultdict.keys():
        resultdict[label] = resultdict[label][idx]
    return resultdict

def process_peptides(args):
    fname = args['file']
    fdr = args['fdr'] / 100
    try:
        outpath = args['outpath']
    except:
        outpath = False

    def calc_sf_all(v, n, p):
        sf_values = -binom.logsf(v, n, p)
        sf_values[np.isinf(sf_values)] = 1
        return sf_values

    elude_path = args['elude']
    elude_path = elude_path.strip()

    ms1results = []
    peps = utils.peptide_gen(args)
    kwargs = prepare_peptide_processor(fname, args)
    func = peptide_processor_iter_isoforms
    print 'Running the search ...'
    for y in utils.multimap(1, func, peps, **kwargs):
        for result in y:
            if len(result):
                ms1results.extend(result)

    prefix = args['prefix']
    protsN, pept_prot, protsSeq = utils.get_prot_pept_map(args)

    resdict = get_results(ms1results)
    del ms1results

    isdecoy = lambda x: x[0].startswith(prefix)
    isdecoy_key = lambda x: x.startswith(prefix)
    escore = lambda x: -x[1]

    p1 = set(resdict['seqs'])

    if len(p1):
        prots_spc2 = defaultdict(set)
        for pep, proteins in pept_prot.iteritems():
            if pep in p1:
                for protein in proteins:
                    prots_spc2[protein].add(pep)

        for k in protsN:
            if k not in prots_spc2:
                prots_spc2[k] = set([])
        prots_spc = dict((k, len(v)) for k, v in prots_spc2.iteritems())

        prots_spc3 = dict()
        for k, v in prots_spc2.iteritems():
            prots_spc3[k] = Counter()
            for pep in v:
                prots_spc3[k][len(pep)] += 1

        prots_spc4 = dict()
        for k, v in protsSeq.iteritems():
            prots_spc4[k] = Counter()
            for pep in v:
                prots_spc4[k][pep] += 1

        names_arr = np.array(prots_spc.keys())
        v_arr = np.array(prots_spc.values())
        n_arr = np.array([protsN[k] for k in prots_spc])

        # top100decoy_score = [prots_spc.get(dprot, 0) for dprot in protsN if isdecoy_key(dprot)]
        # top100decoy_N = [val for key, val in protsN.items() if isdecoy_key(key)]
        # p = np.mean(top100decoy_score) / np.mean(top100decoy_N)
        # print 'p=%s' % (np.mean(top100decoy_score) / np.mean(top100decoy_N))
        p_dict = dict()
        for i in xrange(5, 31, 1):
            top100decoy_score = np.sum([prots_spc3[dprot][i] for dprot in protsN if isdecoy_key(dprot)])
            top100decoy_N = np.sum([prots_spc4[dprot][i] for dprot in protsN if isdecoy_key(dprot)])
            if top100decoy_N > 0:
                p_dict[i] = float(top100decoy_score) / top100decoy_N
            else:
                print('Missing!!! ', i)
        # for k, v in p_dict.iteritems():
        #     print(k, v)
                      
        p_mean = np.mean(p_dict.values())
        p_arr = np.array([p_mean] * len(v_arr))

        p_arr = []
        for dbname in names_arr:
            p_arr.append(np.mean([p_dict.get(ii) for ii in protsSeq[dbname]]))
        p_arr = np.array(p_arr)
        print(np.mean(p_arr))

        p = p_mean
        print(p_mean)

        prots_spc = dict()
        all_pvals = calc_sf_all(v_arr, n_arr, p_arr)
        for idx, k in enumerate(names_arr):
            prots_spc[k] = all_pvals[idx]

        checked = set()
        for k, v in prots_spc.items():
            if k not in checked:
                if isdecoy_key(k):
                    if prots_spc.get(k.replace(prefix, ''), -1e6) > v:
                        del prots_spc[k]
                        checked.add(k.replace(prefix, ''))
                else:
                    if prots_spc.get(prefix + k, -1e6) > v:
                        del prots_spc[k]
                        checked.add(prefix + k)

        filtered_prots = aux.filter(prots_spc.items(), fdr=fdr, key=escore, is_decoy=isdecoy, remove_decoy=True, formula=1,
                                    full_output=True)

        identified_proteins = 0

        for x in filtered_prots:
            identified_proteins += 1
        print 'results for default search: number of identified proteins = %d' % (identified_proteins, )

        print 'Running mass recalibration...'

        true_md = []
        true_isotopes = []
        true_seqs = []
        true_prots = set(x[0] for x in filtered_prots)
        for pep, proteins in pept_prot.iteritems():
            if any(protein in true_prots for protein in proteins):
                true_seqs.append(pep)

        e_ind = np.in1d(resdict['seqs'], true_seqs)

        true_seqs = resdict['seqs'][e_ind]
        true_md.extend(resdict['md'][e_ind])
        true_md = np.array(true_md)
        true_isotopes.extend(resdict['Isotopes'][e_ind])
        true_isotopes = np.array(true_isotopes)

        e_ind = true_isotopes >= 4
        true_md = true_md[e_ind]
        true_seqs = true_seqs[e_ind]

        mass_left = args['ptol']
        mass_right = args['ptol']

        def noisygaus(x, a, x0, sigma, b):
            return a * exp(-(x - x0) ** 2 / (2 * sigma ** 2)) + b

        def calibrate_mass(bwidth, mass_left, mass_right, true_md):

            bbins = np.arange(-mass_left, mass_right, bwidth)
            H1, b1 = np.histogram(true_md, bins=bbins)
            b1 = b1 + bwidth
            b1 = b1[:-1]


            popt, pcov = curve_fit(noisygaus, b1, H1, p0=[1, np.median(true_md), 1, 1])
            mass_shift, mass_sigma = popt[1], abs(popt[2])
            return mass_shift, mass_sigma, pcov[0][0]

        mass_shift, mass_sigma, covvalue = calibrate_mass(0.001, mass_left, mass_right, true_md)
        if np.isinf(covvalue):
            mass_shift, mass_sigma, covvalue = calibrate_mass(0.01, mass_left, mass_right, true_md)
        print 'Calibrated mass shift: ', mass_shift
        print 'Calibrated mass sigma in ppm: ', mass_sigma

        e_all = abs(resdict['md'] - mass_shift) / (mass_sigma)
        r = 3.0
        e_ind = e_all <= r
        resdict = filter_results(resdict, e_ind)

        zs_all = e_all[e_ind] ** 2

        p1 = set(resdict['seqs'])

        prots_spc2 = defaultdict(set)
        for pep, proteins in pept_prot.iteritems():
            if pep in p1:
                for protein in proteins:
                    prots_spc2[protein].add(pep)

        for k in protsN:
            if k not in prots_spc2:
                prots_spc2[k] = set([])
        prots_spc = dict((k, len(v)) for k, v in prots_spc2.iteritems())

        prots_spc3 = dict()
        for k, v in prots_spc2.iteritems():
            prots_spc3[k] = Counter()
            for pep in v:
                prots_spc3[k][len(pep)] += 1
        names_arr = np.array(prots_spc.keys())
        v_arr = np.array(prots_spc.values())
        n_arr = np.array([protsN[k] for k in prots_spc])

        # top100decoy_score = [prots_spc.get(dprot, 0) for dprot in protsN if isdecoy_key(dprot)]
        # top100decoy_N = [val for key, val in protsN.items() if isdecoy_key(key)]
        # p = np.mean(top100decoy_score) / np.mean(top100decoy_N)
        # print 'p=%s' % (np.mean(top100decoy_score) / np.mean(top100decoy_N))
        p_dict = dict()
        for i in xrange(5, 31, 1):
            top100decoy_score = np.sum([prots_spc3[dprot][i] for dprot in protsN if isdecoy_key(dprot)])
            top100decoy_N = np.sum([prots_spc4[dprot][i] for dprot in protsN if isdecoy_key(dprot)])
            if top100decoy_N > 0:
                p_dict[i] = float(top100decoy_score) / top100decoy_N
            else:
                print('Missing!!! ', i)


        p_mean = np.mean(p_dict.values())
        p_arr = np.array([p_mean] * len(v_arr))

        p_arr = []
        for dbname in names_arr:
            p_arr.append(np.mean([p_dict.get(ii) for ii in protsSeq[dbname]]))
        p_arr = np.array(p_arr)
        print(np.mean(p_arr))

        p = p_mean
        print(p_mean)

        prots_spc = dict()
        all_pvals = calc_sf_all(v_arr, n_arr, p_arr)
        for idx, k in enumerate(names_arr):
            prots_spc[k] = all_pvals[idx]

        checked = set()
        for k, v in prots_spc.items():
            if k not in checked:
                if isdecoy_key(k):
                    if prots_spc.get(k.replace(prefix, ''), -1e6) > v:
                        del prots_spc[k]
                        checked.add(k.replace(prefix, ''))
                else:
                    if prots_spc.get(prefix + k, -1e6) > v:
                        del prots_spc[k]
                        checked.add(prefix + k)

        filtered_prots = aux.filter(prots_spc.items(), fdr=fdr, key=escore, is_decoy=isdecoy, remove_decoy=True, formula=1,
                                    full_output=True)

        identified_proteins = 0

        for x in filtered_prots:
            identified_proteins += 1
        print 'results for default search after mass calibration: number of identified proteins = %d' % (identified_proteins, )



        print 'Running RT prediction...'
        true_seqs = []
        true_rt = []
        true_isotopes = []
        true_prots = set(x[0] for x in filtered_prots)#[:5])
        for pep, proteins in pept_prot.iteritems():
            if any(protein in true_prots for protein in proteins):
                true_seqs.append(pep)
        e_ind = np.in1d(resdict['seqs'], true_seqs)


        true_seqs = resdict['seqs'][e_ind]
        true_rt.extend(resdict['rt'][e_ind])
        true_rt = np.array(true_rt)
        true_isotopes.extend(resdict['Isotopes'][e_ind])
        true_isotopes = np.array(true_isotopes)

        e_ind = true_isotopes >= 4
        true_seqs = true_seqs[e_ind]
        true_rt = true_rt[e_ind]
        true_isotopes = true_isotopes[e_ind]

        mcs = np.array([parser.num_sites(seq, parser.expasy_rules['trypsin']) for seq in true_seqs])
        e_ind = mcs == 0
        true_seqs = true_seqs[e_ind]
        true_rt = true_rt[e_ind]
        true_isotopes = true_isotopes[e_ind]

        best_seq = defaultdict(list)
        newseqs = []
        newRTs = []
        for seq, RT in zip(true_seqs, true_rt):
            best_seq[seq].append(RT)
        for k, v in best_seq.items():
            newseqs.append(k)
            newRTs.append(np.median(v))
        true_seqs = np.array(newseqs)
        true_rt = np.array(newRTs)

        RC, outmask = get_RCs2(true_seqs, true_rt)

        if elude_path:
            outtrain = tempfile.NamedTemporaryFile(suffix='.txt')
            outres = tempfile.NamedTemporaryFile(suffix='.txt')
            outres_name = outres.name
            outres.close()
            ns = true_seqs[~outmask]
            nr = true_rt[~outmask]
            ll = len(ns)
            for seq, RT in zip(ns[:ll], nr[:ll]):
                outtrain.write(seq + '\t' + str(RT) + '\n')
            outtrain.flush()

            subprocess.call([elude_path, '-t', outtrain.name, '-e', outtrain.name, '-a', '-g', '-o', outres_name])
            pepdict = dict()
            train_RT = []
            train_seq = []
            for x in open(outres_name).readlines()[3:]:
                seq, RT, RTexp = x.strip().split('\t')
                pepdict[seq] = float(RT)
                train_seq.append(seq)
                train_RT.append(float(RTexp))
            train_RT = np.array(train_RT)
            RT_pred = np.array([pepdict[s] for s in train_seq])
            aa, bb, RR, ss = aux.linear_regression(RT_pred, train_RT)
        else:
            RC = achrom.get_RCs_vary_lcp(true_seqs[~outmask], true_rt[~outmask])
            RT_pred = np.array([achrom.calculate_RT(s, RC) for s in true_seqs])
            aa, bb, RR, ss = aux.linear_regression(RT_pred, true_rt)
        print aa, bb, RR, ss


        best_sigma = ss
        RT_sigma = best_sigma

    else:
        print 'No matches found'

    p1 = set(resdict['seqs'])

    pepdict = dict()    
    if elude_path:
        outtest = tempfile.NamedTemporaryFile(suffix='.txt')
        for seq in p1:
            outtest.write(seq + '\n')
        outtest.flush()

        subprocess.call([elude_path, '-t', outtrain.name, '-e', outtest.name, '-a', '-o', outres_name])
        for x in open(outres_name).readlines()[3:]:
            seq, RT = x.strip().split('\t')
            pepdict[seq] = float(RT)
        outtest.close()
        outtrain.close()
    else:
        for seq in p1:
            pepdict[seq] = achrom.calculate_RT(seq, RC)
    rt_pred = np.array([pepdict[s] for s in resdict['seqs']])
    rt_diff = resdict['rt'] - rt_pred
    e_all = (rt_diff) ** 2 / (RT_sigma ** 2)
    zs_all = zs_all + e_all
    r = 3.0
    e_ind = e_all <= r
    resdict = filter_results(resdict, e_ind)
    rt_diff = rt_diff[e_ind]
    zs_all = zs_all[e_ind]

    e_ind = zs_all <= args['rtt']# ** 2
    resdict = filter_results(resdict, e_ind)
    rt_diff = rt_diff[e_ind]
    zs_all = zs_all[e_ind]


    if outpath:
        base_out_name = os.path.splitext(os.path.join(outpath, os.path.basename(fname)))[0]
    else:
        base_out_name = os.path.splitext(fname)[0]

    with open(base_out_name + '_protsN.csv', 'w') as output:
        output.write('dbname\ttheor peptides\n')
        for k, v in protsN.items():
            output.write('\t'.join((k, str(v))) + '\n')
   
    with open(base_out_name + '_PFMs.csv', 'w') as output:
        output.write('sequence\tmass diff\tRT diff\tpeak_id\tIntensity\tnScans\tnIsotopes\tproteins\tm/z\tRT\taveragineCorr\tcharge\n')
        for seq, md, rtd, peak_id, I, nScans, nIsotopes, mzr, rtr, av, ch in zip(resdict['seqs'], resdict['md'], rt_diff, resdict['ids'], resdict['Is'], resdict['Scans'], resdict['Isotopes'], resdict['mzraw'], resdict['rt'], resdict['av'], resdict['ch']):
            output.write('\t'.join((seq, str(md), str(rtd), str(peak_id), str(I), str(nScans), str(nIsotopes), ';'.join(pept_prot[seq]), str(mzr), str(rtr), str(av), str(ch))) + '\n')
            
    p1 = set(resdict['seqs'])

    pep_pid = defaultdict(set)
    pid_pep = defaultdict(set)
    for pep, pid in zip(resdict['seqs'], resdict['ids']):
        pep_pid[pep].add(pid)
        pid_pep[pid].add(pep)

    if len(p1):
        prots_spc_final = dict()
        prots_spc_copy = False
        banned_dict = dict()
        prots_spc2 = False
        unstable_prots = set()
        p0 = False

        while(len(p1)):
            if not prots_spc2:
                prots_spc2 = defaultdict(set)
                for pep, proteins in pept_prot.iteritems():
                    if pep in p1:
                        for protein in proteins:
                            prots_spc2[protein].add(pep)

                for k in protsN:
                    if k not in prots_spc2:
                        prots_spc2[k] = set([])
                unstable_prots = set(prots_spc2.keys())
            tmp_spc_new = dict((k, sum(banned_dict.get(l, 1) for l in v) if k in unstable_prots else tmp_spc.get(k, 0)) for k, v in prots_spc2.iteritems())
            tmp_spc = tmp_spc_new
            prots_spc = tmp_spc_new
            if not prots_spc_copy:
                prots_spc_copy = deepcopy(prots_spc)


            prots_spc3 = dict()
            for k, v in prots_spc2.iteritems():
                prots_spc3[k] = Counter()
                for pep in v:
                    if banned_dict.get(pep, 1):
                        prots_spc3[k][len(pep)] += 1

            names_arr = np.array(prots_spc.keys())
            # print len(v_arr), np.sum(v_arr == np.array(prots_spc.values()))#sum(zzz1 == zzz2 for zzz1, zzz2 in zip(v_arr, np.array(prots_spc.values())))
            v_arr = np.array(prots_spc.values())
            n_arr = np.array([protsN[k] for k in prots_spc])

            # prots_spc_copy = copy(prots_spc)
            if not p0:
                top100decoy_score = [prots_spc.get(dprot, 0) for dprot in protsN if isdecoy_key(dprot)]
                top100decoy_N = [val for key, val in protsN.items() if isdecoy_key(key)]
                p = np.mean(top100decoy_score) / np.mean(top100decoy_N)
                p0 = float(p)
            # print 'p=%s' % (np.mean(top100decoy_score) / np.mean(top100decoy_N))
            p_dict = dict()
            for i in xrange(5, 31, 1):
                top100decoy_score = np.sum([prots_spc3[dprot][i] for dprot in protsN if isdecoy_key(dprot)])
                top100decoy_N = np.sum([prots_spc4[dprot][i] for dprot in protsN if isdecoy_key(dprot)])
                if top100decoy_N > 0:
                    p_dict[i] = float(top100decoy_score) / top100decoy_N
                else:
                    p_dict[i] = p_dict[i-1]
                    print('Missing!!! ', i)
            # for k, v in p_dict.iteritems():
            #     print(k, v)
                        
            p_mean = np.mean(p_dict.values())
            p_arr = np.array([p_mean] * len(v_arr))

            p_arr = []
            for dbname in names_arr:
                p_arr.append(np.mean([p_dict.get(ii) for ii in protsSeq[dbname]]))
            p_arr = np.array(p_arr)

            p = p_mean

            prots_spc = dict()
            all_pvals = calc_sf_all(v_arr, n_arr, p_arr)
            for idx, k in enumerate(names_arr):
                prots_spc[k] = all_pvals[idx]

            sortedlist_spc = sorted(prots_spc.iteritems(), key=operator.itemgetter(1))[::-1]
            best_prot, best_score = sortedlist_spc[0][0], sortedlist_spc[0][1]
            unstable_prots = set()
            if best_prot not in prots_spc_final:
                prots_spc_final[best_prot] = best_score
                banned_pids = set()
                for pep in prots_spc2[best_prot]:
                    for pid in pep_pid[pep]:
                        banned_pids.add(pid)
                for pid in banned_pids:
                    for pep in pid_pep[pid]:
                        # banned_peptides.add(pep)
                        banned_dict[pep] = 0
                        for bprot in pept_prot[pep]:
                            unstable_prots.add(bprot)
                del banned_pids
            else:
                break

            prot_fdr = aux.fdr(prots_spc_final.items(), is_decoy=isdecoy)
            # print len(prots_spc_final), len(p1), prot_fdr
            num_tot_prots = len(prots_spc_final)
            if num_tot_prots % 25 == 0:
                print 'Approximately %d proteins identified at %.1f %% FDR' % (num_tot_prots, prot_fdr * 100)
                print p
                print(np.mean(p_arr))
                print(p_mean)
            if prot_fdr >= 2.5 * fdr:
                break

                
        print 'p=%s' % (p0, )




        prots_spc = prots_spc_final
        sortedlist_spc = sorted(prots_spc.iteritems(), key=operator.itemgetter(1))[::-1]
        with open(base_out_name + '_proteins_full.csv', 'w') as output:
            output.write('dbname\tscore\tmatched peptides\ttheoretical peptides\n')
            for x in sortedlist_spc:
                output.write('\t'.join((x[0], str(x[1]), str(prots_spc_copy[x[0]]), str(protsN[x[0]]))) + '\n')

        checked = set()
        for k, v in prots_spc.items():
            if k not in checked:
                if isdecoy_key(k):
                    if prots_spc.get(k.replace(prefix, ''), -1e6) > v:
                        del prots_spc[k]
                        checked.add(k.replace(prefix, ''))
                else:
                    if prots_spc.get(prefix + k, -1e6) > v:
                        del prots_spc[k]
                        checked.add(prefix + k)

        filtered_prots = aux.filter(prots_spc.items(), fdr=fdr, key=escore, is_decoy=isdecoy, remove_decoy=True, formula=1, full_output=True)

        identified_proteins = 0

        for x in filtered_prots:
            identified_proteins += 1

        print 'TOP 5 identified proteins:'
        print 'dbname\tscore\tnum matched peptides\tnum theoretical peptides'
        for x in filtered_prots[:5]:
            print '\t'.join((str(x[0]), str(x[1]), str(int(prots_spc_copy[x[0]])), str(protsN[x[0]])))
        print 'results:%s;number of identified proteins = %d' % (fname, identified_proteins, )
        print 'R=', r
        with open(base_out_name + '_proteins.csv', 'w') as output:
            output.write('dbname\tscore\tmatched peptides\ttheoretical peptides\n')
            for x in filtered_prots:
                output.write('\t'.join((x[0], str(x[1]), str(prots_spc_copy[x[0]]), str(protsN[x[0]]))) + '\n')


        fig = plt.figure(figsize=(16, 12))
        DPI = fig.get_dpi()
        fig.set_size_inches(2000.0/float(DPI), 2000.0/float(DPI))

        df0 = pd.read_table(os.path.splitext(fname)[0].replace('.features', '') + '.features' + '.tsv')

        # Features RT distribution
        # TODO add matched features and matched to 1% FDR proteins features
        ax = fig.add_subplot(3, 1, 1)
        bns = np.arange(0, df0['rtApex'].max() + 1, 1)
        ax.hist(df0['rtApex'], bins = bns)
        ax.set_xlabel('RT, min', size=16)
        ax.set_ylabel('# features', size=16)

        # Features mass distribution

        # TODO add matched features and matched to 1% FDR proteins features
        ax = fig.add_subplot(3, 1, 2)
        bns = np.arange(0, df0['massCalib'].max() + 6, 5)
        ax.hist(df0['massCalib'], bins = bns)
        ax.set_xlabel('neutral mass, Da', size=16)
        ax.set_ylabel('# features', size=16)

        # Features intensity distribution

        # TODO add matched features and matched to 1% FDR proteins features
        ax = fig.add_subplot(3, 1, 3)
        bns = np.arange(np.log10(df0['intensityApex'].min()) - 0.5, np.log10(df0['intensityApex'].max()) + 0.5, 0.5)
        ax.hist(np.log10(df0['intensityApex']), bins = bns)
        ax.set_xlabel('log10(Intensity)', size=16)
        ax.set_ylabel('# features', size=16)

        plt.savefig(base_out_name + '.png')

    else:
        print 'No matches found'
